"""Module to test api in a generic way."""
import hashlib
import json
import re

import inflection
import pytest
from django.db import connection
from django.db.models.fields.related import ManyToManyDescriptor
from django.test.utils import CaptureQueriesContext
from django.urls import reverse
from rest_framework_json_api.renderers import JSONRenderer

from ..views import (
    ACLViewSet,
    PermissionViewSet,
    RoleViewSet,
    ScopeViewSet,
    UserViewSet,
)

# @pytest.fixture(params=["admin_client", "authenticated_client"])
# def api_client(db, request):
#     client = request.getfixturevalue(request.param)
#     return client


@pytest.fixture()
def deterministic_uuids(mocker):
    # md5 hex digests are the same length as UUIDs, so django happily accepts
    # them.  Note also this has no cryptographic use here, so md5 is totally
    # fine
    hashy = hashlib.md5()

    def next_uuid():
        # add some string to the hash. This modifies it enough to yield a
        # totally different value
        hashy.update(b"x")
        # format: 9336ebf2-5087-d91c-818e-e6e9 ec29 f8c1
        # lengths: 8        4    4    4    12
        digest = hashy.hexdigest()
        return "-".join(
            [
                digest[:8],  # 8
                digest[8:12],  # 4
                digest[12:16],  # 4
                digest[16:20],  # 4
                digest[20:],  # 4
            ]
        )

    mocker.patch("uuid.uuid4", next_uuid)


@pytest.fixture(
    params=[
        # add your viewset and expected queries run for generic api testing...
        ACLViewSet,
        PermissionViewSet,
        RoleViewSet,
        ScopeViewSet,
        UserViewSet,
    ]
)
def viewset(request):
    """
    Viewset fixture listing viewsets for generic testing.

    For generic testing viewset needs to meet following requirements:
    * class fields `serializer_class` and `queryset`
    * registered factory for model it serves
    * registered uri with `SimpleRouter`
    """
    viewset_instance = request.param()
    model = viewset_instance.queryset.model
    instance_name = inflection.underscore(model._meta.object_name)
    base_name = model._meta.object_name.lower()

    viewset_instance.factory_name = f"{instance_name}_factory"
    viewset_instance.instance_name = instance_name
    viewset_instance.base_name = base_name
    viewset_instance.kwargs = {}

    return viewset_instance


@pytest.fixture
def fixture(
    deterministic_uuids,
    django_db_reset_sequences,
    request,
    viewset,
):
    """Get fixture and many to many relations of given viewset."""
    fixture = request.getfixturevalue(viewset.factory_name)()
    if hasattr(viewset.serializer_class, "included_serializers"):
        for name in sorted(viewset.serializer_class.included_serializers.keys()):
            relation_type = getattr(fixture.__class__, name)
            # pytest factory boy doesn't have native ManyToMany support
            # so needs to be handled manually
            if isinstance(relation_type, ManyToManyDescriptor):
                request.getfixturevalue("{0}_{1}".format(viewset.instance_name, name))

    return fixture


def prefetch_query_in_normalizer(query):
    """
    Normalize `IN` queries.

    Using `prefetch_related()` leads to `IN` queries where we have no control over the
    order of the parameters. Make sure the order in the snapshot is always
    alphabetical.
    """
    regex = r".* IN \((.*)\).*"
    match = re.match(regex, query)
    if match and match.groups():
        for group in match.groups():
            lst = group.split(", ")
            new_string = ", ".join(sorted(lst))
            query = query.replace(group, new_string)
    return query


def assert_response(response, query_context, snapshot, include_json=True):
    value = {
        "status": response.status_code,
        "queries": [
            prefetch_query_in_normalizer(query["sql"])
            for query in query_context.captured_queries
        ],
        "request": {
            k: v for k, v in response.request.items() if not k.startswith("wsgi")
        },
    }

    # Drop `SAVEPOINT` statements because they will change on every run.
    value["queries"] = list(
        filter(lambda lst: "SAVEPOINT" not in lst, value["queries"])
    )

    if include_json:
        value["response"] = response.json()

    snapshot.assert_match(value)


@pytest.mark.freeze_time("2017-05-21")
def test_api_list(fixture, request, admin_client, snapshot, viewset):
    url = reverse("{0}-list".format(viewset.base_name))

    # create data for proper num queries check
    request.getfixturevalue(viewset.factory_name).create_batch(2)

    included = getattr(viewset.serializer_class, "included_serializers", {})
    with CaptureQueriesContext(connection) as context:
        response = admin_client.get(url, data={"include": ",".join(included.keys())})

    assert_response(response, context, snapshot)


@pytest.mark.freeze_time("2017-05-21")
def test_api_detail(fixture, admin_client, viewset, snapshot):
    url = reverse("{0}-detail".format(viewset.base_name), args=[fixture.pk])

    included = getattr(viewset.serializer_class, "included_serializers", {})
    with CaptureQueriesContext(connection) as context:
        response = admin_client.get(url, data={"include": ",".join(included.keys())})

    assert_response(response, context, snapshot)


@pytest.mark.freeze_time("2017-05-21")
def test_api_create(transactional_db, fixture, admin_client, viewset, snapshot):
    url = reverse("{0}-list".format(viewset.base_name))

    serializer = viewset.serializer_class(fixture)
    renderer = JSONRenderer()
    context = {"view": viewset}
    data = renderer.render(serializer.data, renderer_context=context)
    fixture.delete()  # avoid constraint issues

    with CaptureQueriesContext(connection) as context:
        response = admin_client.post(url, data=json.loads(data))

    assert_response(response, context, snapshot)


@pytest.mark.freeze_time("2017-05-21")
def test_api_patch(fixture, admin_client, viewset, snapshot):
    url = reverse("{0}-detail".format(viewset.base_name), args=[fixture.pk])

    serializer = viewset.serializer_class(fixture)
    renderer = JSONRenderer()
    context = {"view": viewset}
    data = renderer.render(serializer.data, renderer_context=context)

    with CaptureQueriesContext(connection) as context:
        response = admin_client.patch(url, data=json.loads(data))

    assert_response(response, context, snapshot)


@pytest.mark.freeze_time("2017-05-21")
def test_api_destroy(fixture, admin_client, snapshot, viewset):
    url = reverse("{0}-detail".format(viewset.base_name), args=[fixture.pk])

    with CaptureQueriesContext(connection) as context:
        response = admin_client.delete(url)

    assert_response(response, context, snapshot, False)



from tempfile import NamedTemporaryFile

import openpyxl
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from generic_permissions.permissions import PermissionViewMixin
from generic_permissions.visibilities import VisibilityViewMixin
from rest_framework.decorators import action
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.permissions import IsAuthenticated
from rest_framework.viewsets import GenericViewSet
from rest_framework_json_api import views

from . import filters, models, serializers


class BaseViewset(VisibilityViewMixin, PermissionViewMixin, views.ModelViewSet):
    pass


class MeViewSet(RetrieveModelMixin, GenericViewSet):
    """Me view returns current user."""

    serializer_class = serializers.MeSerializer
    permission_classes = [IsAuthenticated]

    def get_object(self, *args, **kwargs):
        return self.request.user


class MyACLViewSet(RetrieveModelMixin, ListModelMixin, GenericViewSet):
    """MyACL view returns current users ACLs."""

    queryset = models.ACL.objects.all()
    serializer_class = serializers.MyACLSerializer
    permission_classes = [IsAuthenticated]

    def get_object(self):
        pk = self.kwargs.pop("pk")
        acl = get_object_or_404(models.ACL, pk=pk)
        if acl.user != self.request.user:
            raise PermissionDenied()
        return acl

    def get_queryset(self):
        return self.queryset.filter(user=self.request.user)


class UserViewSet(BaseViewset):
    serializer_class = serializers.UserSerializer
    queryset = models.User.objects.all()

    filterset_class = filters.UserFilterset

    search_fields = (
        "username",
        "first_name",
        "last_name",
        "email",
        "=zip",
        "city",
    )

    case_insensitive_ordering_fields = [
        "first_name",
        "last_name",
        "email",
        "address",
        "city",
    ]

    @action(methods=["get"], detail=False)
    def export(self, request):
        """Export user list as excel table."""
        queryset = self.get_queryset()

        # we use a template to set the headers to bold and align the columns to fit on a DIN-A4 page
        workbook = openpyxl.load_workbook(settings.USER_EXPORT_TEMPLATE_PATH)
        sheet = workbook.active

        # set column headers
        sheet["A1"] = _("Name")
        sheet["B1"] = _("First Name")
        sheet["C1"] = _("Email")
        sheet["D1"] = _("Roles and organizations")

        for row, user in enumerate(queryset.iterator(), start=2):
            acl_string = "\n".join(
                [
                    f"{acl.role.name}: {acl.scope.full_name()}"
                    for acl in user.acls.all()
                    .select_related("scope", "role")
                    .order_by("role__name")
                ]
            )

            sheet[f"A{row}"] = user.last_name
            sheet[f"B{row}"] = user.first_name
            sheet[f"C{row}"] = user.email
            sheet[f"D{row}"] = acl_string
            sheet[f"D{row}"].alignment = openpyxl.styles.Alignment(wrap_text=True)

        # save workbook as temporary file and read the file
        with NamedTemporaryFile() as tmp:
            workbook.save(tmp.name)
            tmp.seek(0)
            stream = tmp.read()
            workbook.close()
            tmp.close()

        response = HttpResponse(stream, content_type="application/vnd.ms-excel")
        response["Content-Disposition"] = "attachment; filename=export.xlsx"
        return response


class ScopeViewSet(BaseViewset):
    serializer_class = serializers.ScopeSerializer
    queryset = models.Scope.objects.all()
    search_fields = (
        "name",
        "description",
    )
    case_insensitive_ordering_fields = [
        "name",
        "description",
    ]


class RoleViewSet(BaseViewset):
    serializer_class = serializers.RoleSerializer
    queryset = models.Role.objects.all()
    search_fields = (
        "slug",
        "name",
        "description",
    )
    case_insensitive_ordering_fields = [
        "name",
        "description",
    ]


class PermissionViewSet(BaseViewset):
    serializer_class = serializers.PermissionSerializer
    queryset = models.Permission.objects.prefetch_related("roles")
    filterset_fields = ("roles",)
    search_fields = (
        "slug",
        "name",
        "description",
    )
    case_insensitive_ordering_fields = [
        "name",
        "description",
    ]


class ACLViewSet(BaseViewset):
    serializer_class = serializers.ACLSerializer
    queryset = models.ACL.objects.all()
    filterset_fields = ("user", "scope", "role")
    search_fields = (
        "scope__name",
        "scope__description",
        "role__name",
        "role__description",
        "user__username",
        "user__first_name",
        "user__last_name",
        "=user__zip",
        "user__city",
    )



import mptt
import mptt.managers


def rebuild_mptt_model(model):
    """
    Rebuild an mptt model.

    This is used in migrations that work with data from mptt models.
    """
    manager = mptt.managers.TreeManager()
    manager.model = model
    mptt.register(model)
    manager.contribute_to_class(model, "objects")
    manager.rebuild()



"""Setuptools package definition."""

import os

from setuptools import find_packages, setup

version = {}
with open("emeis/emeis_metadata.py") as fp:
    exec(fp.read(), version)


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name=version["__title__"],
    version=version["__version__"],
    description=version["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="projectcaluma",
    url="https://github.com/projectcaluma/emeis",
    license="License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    install_requires=[
        "django~=3.2",
        "django-cors-headers >=3.7.0, <=3.13.1",
        "django-environ==0.8.1",
        "django-filter==21.1",
        "django-generic-api-permissions==0.2.0",
        "django-localized-fields>=6.4,<=6.6",
        "django_mptt>=0.11,<=0.13.4",
        "django-postgres-extra==2.0.3",
        "djangorestframework>=3.12.4,<=3.13.1",
        "djangorestframework-jsonapi>=4.3.0,<=5.0.0",
        "mozilla-django-oidc==1.2.4",
        "pyexcel==0.6.7",
        "pyexcel-xlsx==0.6.0",
        "requests==2.26.0",
        "uwsgi==2.0.19.1",
        "openpyxl==3.0.9",
    ],
)
